#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Lightweight annotations to classes and methods for descriptive purposes.

These do not actually do anything useful other than appending to the object's docstring, they just make the code
easier to self document in a consistent way without using hungarian-notation-like names.

Note:
    Remember that hints are purely descriptive notes added to the docstring. They must not called
    with arguments, for example, you would use `@hint`. Calling *with* parenthesis `()` is invalid.

    :mod:`metabeyond.remarks` decorators on the other hand must be invoked with an argument list. An example being
    `@remark()` or `@remark(foo, bar)` or `@remark(f='foo', b='bar')`. These will add metadata to the decorated
    function or class that specifies the presence of that decorator on the object, and this can be queried at runtime.
    Calling *without* parenthesis will produce an invalid result.

    Using the presence of parenthesis is a good way to determine if it is a documentation hint or a
    metadata remark. Metadata remarks are much more complicated than hints.

    To document parameter types and return types, one should use type annotations or comment-based
    annotations as per PEP 3107: `def foo(bar: int, baz: int) -> float:`. Utilities for this are in the
    Python :mod:`typing` module.

Warning:
    Enabling CPython's second level optimisation `-OO` will prevent docstrings from being retained by classes and
    functions. Applying decorators will force those docstrings to be implemented, so it will appear that the other
    contents will have been replaced by the hint docstring. This is due to how Python handles optimisation, and it is
    out of scope of this module to take this into consideration.

    TL;DR don't run Sphinx with optimisation flags enabled, and if you reflect on your docstrings at runtime, don't
    use `-OO`. Usage of first level optimisation `-O` is fine.
"""
import textwrap
from typing import Any, Callable, Type, Union

#: Type hint for valid things to decorate.
_ToDecorate = Union[Callable[..., Any], Type[Any]]


def append_to_docstring(new_content: str, existing_object: Any) -> None:
    """
    Appends a given string to the docstring of an existing object. If no docstring exists, one is injected.

    Will fail on frozen objects, such as C modules.

    Will fail if the docstring is a non-string or non-None type.
    """
    existing_doc = getattr(existing_object, "__doc__", "")
    if existing_doc is None:
        existing_doc = ""

    elif not isinstance(existing_doc, str):
        raise TypeError(
            f"{existing_object} was expected to have a string docstring, but it was {existing_doc}"
        )

    # Dedent to add the documentation, then reindent again to maintain formatting.
    start_index = existing_doc.index(existing_doc.lstrip())
    indent = existing_doc[:start_index]
    existing_doc = textwrap.dedent(existing_doc)

    if existing_doc and not existing_doc.isspace():
        existing_doc += "\n"

    existing_doc += new_content
    existing_doc = textwrap.indent(existing_doc, indent, lambda _: None)
    setattr(existing_object, "__doc__", existing_doc)


class Hint:
    """
    Simple hint base implementation which produces a decorator object with some description.

    The decorator always returns what was input without any modification other than appending to the docstring.
    This mechanism is purely for documentation purposes only.

        >>> interface = Hint('This is an abstract class with no internal state.')

        >>> @interface
        ... class Playable:
        ...     "A thing that you can play."
        ...     def play(self):
        ...         ...

        >>> import inspect
        >>> print(inspect.cleandoc(inspect.getdoc(Playable)))

    The above example would produce the following

    .. code-block:: none

        A thing that you can play.

        This is an abstract class with no internal state.

    Hints implement a custom `__repr__` method to allow them to self-document
    in a Sphinx environment.
    """

    def __init__(self, description: str) -> None:
        self.description = description

    def __call__(self, item):
        append_to_docstring(self.description, item)
        return item

    def __repr__(self):
        return "This hint reads: " + self.description


#: Decorates anything used as a decorator.
decorator = Hint("This is a class or function used as a decorator.")

#: Decorates anything that behaves like an interface with no underlying state.
interface = Hint("This is a pure abstract class with no underlying state.")

#: Decorates anything that might be a placeholder forward declaration.
forward_declaration = Hint("This is a forward declaration for another object")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Remarks for marking a method in a remark as a predicate or constraint.
"""
__all__ = ["RemarkPredicate", "RemarkConstraint"]

import inspect


# We cant use remarks to define core remarks as it would lead to a recursive reference and break, so we make some
# cheap modifications here to keep stuff simple.


class RemarkConstraint:
    """
    Decorates a function within a :class:`Remark`-derived class to register the function as a constraint method.

    Note:
        This decorator must not be called with any arguments.

        >>> # Correct
        >>> @RemarkConstraint
        ... def foo(self, item): ...

        >>> # Incorrect
        >>> @RemarkConstraint()
        ... def foo(self, item): ...
    """

    def __init__(self, func):
        if not inspect.isfunction(func):
            raise TypeError("Cannot mark a non-function as a remark constraint...")
        else:
            self.func = func

    def __call__(self, remark, item):
        return self.func(remark, item)


class RemarkPredicate:
    """
    Decorates a function within a :class:`Remark`-derived class to register the function as a predicate method.

    Note:
        This decorator must not be called with any arguments.

        >>> # Correct
        >>> @RemarkPredicate
        ... def foo(self, item): ...

        >>> # Incorrect
        >>> @RemarkPredicate()
        ... def foo(self, item): ...

    """

    def __init__(self, func):
        if not inspect.isfunction(func):
            raise TypeError("Cannot mark a non-function as a remark predicate...")
        else:
            self.func = func

    def __call__(self, remark, item):
        return self.func(remark, item)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implementation of a Remark.
"""
__all__ = ["Remark"]

import collections
import inspect

from metabeyond import hints
from . import _intern
from . import predication


@hints.decorator
class Remark:
    # noinspection PyUnresolvedReferences
    """
    Class that can be subclassed to produce remark decorators that behave similar to Java 6 annotations::

    >>> class Foo(Remark): ...

    >>> @Foo()
    ... def bar():
    ...     pass

    You may also make it accept other parameters.

    >>> class Baz(Remark):
    >>>     def __init__(self, *, x: int, y: int):
    ...         super().__init__()
    ...         self.x = x
    ...         self.y = y

    And decorating with multiple remarks is also supported.

    >>> @Foo()
    ... @Baz(x=122, y=34)  # kwargs work too
    ... def bork():
    ...     pass

    Constraints:

        You can add a single runtime constraint onto a class in order to limit what it is allowed to decorate when used.
        These will be inherited regardless of whether you specify one or not.

        >>> import asyncio
        >>> class SomeRemark(Remark, constraint=asyncio.iscoroutinefunction):
        ...     ...

        >>> @SomeRemark()
        ... async def whatever():
        ...      "This is acceptable!"

        >>> @SomeRemark()
        ... def something_else():
        ...     "This will cause a type error..."

        Or supply multiple constraints in a collection type.

        >>> import asyncio
        >>> import inspect
        >>> class SomeOtherRemark(Remark,
        ...                       constraints=[
        ...                           asyncio.iscoroutinefunction,
        ...                           lambda o: bool(inspect.getdoc(o).strip())
        ...                       ]):
        ...    ...

        Constraints should be callable type `Callable[[ToDecorate], bool]`, meaning they take the decorated
        item as a sole parameter and return True if it is acceptable or False otherwise. They may alternatively
        raise their own exceptions for a more custom interface. Multiple constraints will all have to be True to enable
        the entire check suite to pass.

    Predicates:

        Since v0.1.0, predicates have also been supported in remarks. Predicates are similar to constraints in that they
        are executed across the decorated item when you decorate an object with the remark.

        The distinguishing feature is that whereas constraints will create an error when they fail, a predicate will
        only result in the object not being decorated by the remark if it fails.

        This is useful if you wish to only decorate an object based on some runtime condition, environment variable, or
        flag.

        As the name suggests, a predicate that returns True is considered to be successful, whereas a predicate that
        returns False will be considered to have failed.

        Since v0.2.0, exceptions raised by predicates have been propagated to the caller rather than being interpreted
        as False results.

        Like constraints, a single predicate is provided with the `predicate` kwarg, or multiple predicates can be
        provided with the `predicates` kwarg.

        For example, adding a single predicate:

        >>> import os
        >>> class SomeRemark(Remark, predicate=lambda _: os.getenv('ADD_REMARKS', False)):
        ...    ...

        ... or multiple predicates with the following syntax:

        >>> import os
        >>> class SomeRemark(Remark, predicates=[lambda _: os.getenv('ADD_REMARKS', False), something_else]):
        ...    ...

        This remark would only apply to an object if the `ADD_REMARKS` environment variable was applied at the time
        the object was decorated.

    Warning:
        Passing the `predicate` and `predicates` arguments together in the class definition will cause an error,
        as will `constraint` and `constraints`. This is purely because because I deemed it a stupid thing to be doing.
        You have no need to use both. Make your code reasonable, sensible and readable!

    As of v0.2.0, constraints and predicates may alternatively be specified by defining methods within the class
    with the :attr:`RemarkConstraint` and :attr:`RemarkPredicate` remarks attached to them. This allows defining
    constraints and predicates with the following syntax:

        >>> # Notice how the kwarg constraint will still work
        >>> class SomeRemark(Remark, constraint=inspect.isclass):
        ...     @RemarkConstraint
        ...     def ensure_is_derived_from_foo(self, item_to_annotate):
        ...         return isinstance(item_to_annotate, Foo)
        ...
        ...     @RemarkPredicate
        ...     def only_add_remark_if_var_is_set(self, item_to_annotate):
        ...         return os.getenv('ADD_REMARKS', False)

    The two arguments passed to these methods are `self`, which is the Remark object being called from, and the
    `item_to_annotate` which is the object being decorated by the remark that the predication function is checking.
    """

    def __init__(self) -> None:
        if type(self) is Remark:
            raise TypeError("Please subclass this class first")

    def __init_subclass__(cls, **kwargs) -> None:
        cls._init_constraints(**kwargs)
        cls._init_predicates(**kwargs)

    @classmethod
    def _init_constraints(cls, **kwargs):

        if not hasattr(cls, _intern.CONSTRAINT_ATTR):
            _intern.logger.debug("Applying new constraint mapping to %s", cls)
            constraints = collections.defaultdict(list)
            setattr(cls, _intern.CONSTRAINT_ATTR, constraints)
        else:
            constraints = getattr(cls, _intern.CONSTRAINT_ATTR)

        if _intern.CONSTRAINT_ARG in kwargs and _intern.CONSTRAINTS_ARG in kwargs:
            raise TypeError(
                f"Please specify either {_intern.CONSTRAINTS_ARG} or {_intern.CONSTRAINT_ARG}"
            )
        if _intern.CONSTRAINTS_ARG in kwargs:
            _intern.logger.debug(
                "Adding %s constraints to %s", len(kwargs[_intern.CONSTRAINTS_ARG]), cls
            )
            constraints[cls].extend(kwargs[_intern.CONSTRAINTS_ARG])
        elif _intern.CONSTRAINT_ARG in kwargs:
            _intern.logger.debug("Adding one constraint to %s", cls)
            constraints[cls].append(kwargs[_intern.CONSTRAINT_ARG])

        methods = _intern.members_in_this_class(
            cls, lambda item: isinstance(item, predication.RemarkConstraint)
        )
        constraints[cls].extend(methods)

    @classmethod
    def _init_predicates(cls, **kwargs):
        if not hasattr(cls, _intern.PREDICATE_ATTR):
            _intern.logger.debug("Applying new predicate mapping to %s", cls)
            predicates = collections.defaultdict(list)
            setattr(cls, _intern.PREDICATE_ATTR, predicates)
        else:
            predicates = getattr(cls, _intern.PREDICATE_ATTR)

        if _intern.PREDICATE_ARG in kwargs and _intern.PREDICATES_ARG in kwargs:
            raise TypeError(
                f"Please specify either {_intern.PREDICATES_ARG} or {_intern.PREDICATE_ARG}"
            )
        if _intern.PREDICATES_ARG in kwargs:
            _intern.logger.debug(
                "Adding %s predicates to %s", len(kwargs[_intern.PREDICATES_ARG]), cls
            )
            predicates[cls].extend(kwargs[_intern.PREDICATES_ARG])
        elif _intern.PREDICATE_ARG in kwargs:
            _intern.logger.debug("Adding one predicate to %s", cls)
            predicates[cls].append(kwargs[_intern.PREDICATE_ARG])

        methods = _intern.members_in_this_class(
            cls, lambda item: isinstance(item, predication.RemarkPredicate)
        )
        predicates[cls].extend(methods)

    def __call__(self, item_to_wrap: _intern.ToDecorate) -> _intern.ToDecorate:
        # Apply constraints first.
        self._perform_constraint_checking(item_to_wrap)

        if self._perform_predicate_checking(item_to_wrap):
            if not hasattr(item_to_wrap, _intern.REMARK_COLLECTION_ATTR):
                setattr(item_to_wrap, _intern.REMARK_COLLECTION_ATTR, set())
            getattr(item_to_wrap, _intern.REMARK_COLLECTION_ATTR).add(self)

        return item_to_wrap

    def __repr__(self):
        middle_bit = " ".join(
            f"{k!s}={v!r}" for k, v in self.__dict__.items() if not k.startswith("_")
        )
        return f"{type(self).__name__}({middle_bit})"

    def __hash__(self):
        """
        We hash the class object, this means each class may only be annotated once per remark type.

        This does not take into account transitive inheritance.

        Warning:
             Do not overload this method, otherwise it may cause this framework to misbehave.
        """
        return hash(type(self))

    def __eq__(self, other):
        """
        Compares the class.

        This may seem counter intuitive, but we hash on the class ID anyway. The reason for this is to allow
        a single remark of each type to be allowed on an object only. We must apply an ``__eq__`` overload as
        per https://docs.python.org/3.7/glossary.html#term-hashable

        There is not really a use case for comparing by content value, remark instances usually would be compared
        by their identity with the ``is`` operator instead.

        Warning:
             Do not overload this method, otherwise it may cause this framework to misbehave.
        """
        return hash(self) == hash(other)

    def _perform_constraint_checking(self, item_to_wrap) -> None:
        """
        Checks any constraints on the object and fails if any are not met.
        """
        for enforcing_class, check_list in getattr(self, _intern.CONSTRAINT_ATTR).items():
            for check in check_list:
                if isinstance(check, predication.RemarkConstraint):
                    # Unwrap correctly, as we assume it is an instance method.
                    result = check(self, item_to_wrap)
                else:
                    result = check(item_to_wrap)

                if not result:
                    friendly_item = getattr(item_to_wrap, "__name__", item_to_wrap)
                    friendly_check = getattr(check, "__name__", check)
                    # This error should explain exactly what failed to work...
                    raise TypeError(
                        f'"{type(self).__name__}" cannot be used to decorate "{friendly_item}" because '
                        f'check "{friendly_check}" failed on it, and is required to pass by '
                        f'remark "{enforcing_class.__name__}"'
                    )

    def _perform_predicate_checking(self, item_to_wrap):
        """
        Checks any predicates on this object against the wrapped item. If any
        fail, we return False, else, we return True.
        """
        for enforcing_class, predicate_list in getattr(self, _intern.PREDICATE_ATTR).items():
            for predicate in predicate_list:
                if isinstance(predicate, predication.RemarkPredicate):
                    # Unwrap correctly, as we assume it is an instance method.
                    result = predicate(self, item_to_wrap)
                else:
                    result = predicate(item_to_wrap)

                if not result:
                    _intern.logger.debug(
                        "Predicate %s enforced by %s failed, the object %s will not be marked with "
                        "this remark",
                        predicate,
                        enforcing_class,
                        item_to_wrap,
                    )

                    return False

        else:
            return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Shared elements used internally only. These should not be changed or manipulated at runtime.
"""
import inspect
import logging
from typing import Any, Callable, Type, Union, List

#: Name of the hidden dunder attribute storing remarks.
REMARK_COLLECTION_ATTR = "__metabeyond_remarks__"

#: Frozen set that is reused internally.
# noinspection NonAsciiCharacters
Ø = frozenset()

#: Used for logging debugging information about checks.
logger = logging.getLogger("metabeyond.remarks")

#: Used internally to hold constraints dynamically.
CONSTRAINT_ATTR = "_constraints"

#: Used internally to hold predicates dynamically.
PREDICATE_ATTR = "_predicates"

#: Single constraint
CONSTRAINT_ARG = "constraint"

#: Multiple constraints.
CONSTRAINTS_ARG = "constraints"

#: Single predicate
PREDICATE_ARG = "predicate"

#: Multiple predicates
PREDICATES_ARG = "predicates"

#: Type hint for valid things to decorate.
ToDecorate = Union[Callable[..., Any], Type[Any]]


def members_in_this_class(cls: Type[Any], predicate=lambda _: True) -> List[Any]:
    """
    Performs analysis of the given class object's MRO and extracts members only belonging to the
    class itself, not members inherited from other classes.
    """
    # Maps class in mro to methods they have.
    mro_lookup = dict()

    # Store string names, since some members may not be hashable, we cannot make them into a set.
    for class_ in inspect.getmro(cls):
        mro_lookup[class_] = {name for name, _ in inspect.getmembers(class_)}

    all_members = mro_lookup.pop(cls)

    # Calculate members only in this class.
    for _, member_list in mro_lookup.items():
        all_members -= member_list

    # Resolve to a list of member objects, rather than names, filter by the predicate.
    all_member_objects = [getattr(cls, member) for member in all_members]
    return [member for member in all_member_objects if predicate(member)]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Remark introspection methods.
"""
__all__ = ["quick_remark", "get_remarks_for", "has_remarks", "get_remark", "get_remarks_by_type"]

from typing import Any, FrozenSet, Iterator, Optional, Type

from . import _intern
from .impl import Remark


def quick_remark(name: str) -> Type[Remark]:
    """
    Creates a remark quickly.

    >>> foo = quick_remark('foo')

    >>> @foo()
    ... def bar(): ...

    Args:
        name:
            The name of the remark to use

    Returns:
        The remark type generated from the given name.
    """
    cls: Type[Remark] = type(name, (Remark,), {})
    return cls


def get_remarks_for(object: Any) -> FrozenSet[Remark]:
    """
    Returns a frozen set of remarks for the given object at the time of calling, or an empty set if none exist.

    Args:
        object:
            The object to get the remarks applied to. This might be any type of object, but usually will be a function
            or class reference.

    Returns:
        Frozen set of any remarks found.
    """
    return frozenset(getattr(object, _intern.REMARK_COLLECTION_ATTR, _intern.Ø))


def has_remarks(object: Any) -> bool:
    """
    Returns true if the object has any remarks applied to it.
    """
    return bool(get_remarks_for(object))


def get_remark(class_: Type[Remark], object: Any) -> Optional[Remark]:
    """
    Get the remark of the given type that is applied to the object.

    This will not consider subtypes/covariants.

    Args:
        class_:
            The type of remark to look for.
        object:
            The object with remarks applied to it to inspect.

    Returns:
        The first remark found, or None if no remarks matching this type were discovered.
    """
    remarks = get_remarks_for(object)
    for remark in remarks:
        if type(remark) == class_:
            return remark
    return None


def get_remarks_by_type(base_class: Type[Remark], object: Any) -> Iterator[Remark]:
    """
    Gets any remarks that are instances of the given base class which are applied to the given object.

    Args:
        base_class:
            The base class of remark to filter by. Set to `remark.Remark` to get all remarks, or just
            call :meth:`get_remarks_for`.
        object:
            The object with remarks applied to it to consider.

    Returns:
        An iterator of each matching remark discovered. Order is undefined.
    """
    remarks = get_remarks_for(object)
    for remark in remarks:
        if isinstance(remark, base_class):
            yield remark



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"Annotation" decorators that can be applied to functions and classes as a way of tagging the object with
some form of metadata. This is designed to function in a similar way to Java 6 annotation interfaces.

Note:
    It is imperative to remember that remark decorators hand must be invoked with an argument list. An example being
    `@remark()` or `@remark(foo, bar)` or `@remark(f='foo', b='bar')`. These will add metadata to the decorated
    function or class that specifies the presence of that decorator on the object, and this can be
    queried at runtime, which requires marginally more resources than if we were to not use it at all.
    Calling *without* parenthesis `()` is invalid.

    Hints (provided by :mod:`metabeyond.hints`), on the other hand, are purely descriptive notes added to the docstring.
    They *must not* be called with arguments, for example `@hint`. Calling *with* parenthesis `()` will produce an
    invalid result, not decorating the given object correctly.

    Using the presence of parenthesis is a good way to determine if it is a documentation hint or a
    metadata remark. Metadata remarks are much more complicated than hints.

    To document parameter types and return types, one should use type annotations or comment-based
    annotations as per PEP 3107: `def foo(bar: int, baz: int) -> float:`. Utilities for this are in the
    Python :mod:`typing` module.

As this functionality is fairly difficult to debug due to its dynamic nature, enabling the `DEBUG` logging
level for `metabeyond.remarks` in the standard Python logging module will enable exposing information regarding
what is happening internally, such as the predicates and constraints that are failing, etc.
"""

from .impl import *
from .introspect import *
from .predication import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Useful JSON type aliases.

References:
    - RFC-4627_
    - https://github.com/python/typing/issues/182

.. _RFC-4627: https://tools.ietf.org/html/rfc4627
"""
from typing import Mapping, Optional, Sequence, TypeVar, Union

# Create a woolly basic definition to start, as MyPy doesn't support recursive hints yet.
#: Any valid JSON type that is **not** `null`
AnyJSON = Union["JSONBool", "JSONNumber", "JSONString", "JSONArray", "JSONObject"]

#: Any valid JSON type. This is allowed to be `null`.
OptionalAnyJSON = Optional[AnyJSON]

#: A null value in a JSON document.
JSONNull = None

#: A boolean value in a JSON document.
JSONBool = TypeVar("JSONBool", bound=bool)
#: A :attr:`JSONBool` that may be `null`.
OptionalJSONBool = Optional[JSONBool]

#: A number (int or float) in a JSON document.
JSONNumber = Union[float, int]

#: A :attr:`JSONNumber` that may be `null`.
OptionalJSONNumber = Optional[JSONNumber]

#: A string in a JSON document. This may be interpreted in Python as :class:`str` or :class:`bytes` objects.
JSONString = Union[str, bytes]

#: A :attr:`JSONString` that may be `null`.
OptionalJSONString = Optional[JSONString]

#: An array (sequence of :attr:`AnyJSON` elements) in a JSON document.
JSONArray = Sequence[OptionalAnyJSON]

#: A :attr:`JSONArray` that may be `null`.
OptionalJSONArray = Optional[JSONArray]

#: An object (mapping of non-null :attr:`JSONString` to :attr:`AnyJSON` elements) in a JSON document.
JSONObject = Mapping[JSONString, OptionalAnyJSON]

#: A :attr:`JSONObject` that may be `null`.
OptionalJSONObject = Optional[JSONObject]

#: A valid root element in a JSON document. As per RFC-4627_, this must be a non-`null` :attr:`JSONArray` or
#: :attr:`JSONObject` element.
RootJSON = Union[JSONArray, JSONObject]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Flitt3r"
__copyright__ = f"(c) 2019 {__author__}"
__license__ = "MIT"
__version__ = "0.2.2"
__url__ = "https://gitlab.com/flitt3r/metabeyond"
